MODULE KEYS
  ! keyword parameters that are globally used in many different places in the code
  IMPLICIT NONE

   ! -------- Other ---------------
  CHARACTER*100 :: ACTION
  INTEGER :: RNGSEED
  LOGICAL :: VERBOSE
  DOUBLE PRECISION :: BPLEN 

  ! ----------------------
  ! chain geometry
  ! ---------------------
  INTEGER :: NSEGPERLINK, NBENDREPLIC, NTAIL1SEG, NTAIL2SEG
  DOUBLE PRECISION :: BENDSEP, TWIST
  DOUBLE PRECISION, DIMENSION(3) :: BENDTANM, BENDTANP, POSM, POSP,XAXM,XAXP

  INTEGER, PARAMETER :: MAXFILLBEADS = 1000
  INTEGER :: FILLINBEADS, SPHAND
  DOUBLE PRECISION :: SPRADIUS, SPHEIGHT, SPTWIST, SPTWIST0, SPLEN
  DOUBLE PRECISION :: FILLBEADLIST(MAXFILLBEADS, 6)
  LOGICAL :: SPIRALBENDS

  LOGICAL :: USEFIXDIAM, FIXHELIXPARAM(6)
  DOUBLE PRECISION :: FIXDIAMETER
  DOUBLE PRECISION :: SAMEECUT, SAMEDISTCUT

  LOGICAL :: STARTFROMHELIX, HELCRDSET(6)
  DOUBLE PRECISION :: HELH, HELT, HELR, HELA, HELB, HELG

  ! ---------------
  ! energy parameters
  ! ---------------
  INTEGER, PARAMETER :: MAXNCHARGE = 1000
  CHARACTER*100 :: CYLINDERFILE, SEGCYLFILE, NUCSEGCYLFILE
  LOGICAL :: CYLINDERFROMFILE, SEGCYLFROMFILE, NUCSEGCYLFROMFILE
  INTEGER :: CYLINDERPTS(3)
  DOUBLE PRECISION :: ESTERIC, LP, LSTRETCH, LTW
  LOGICAL :: BENDSTERICS, SEGSTERICS, BENDSEGSTERICS
  DOUBLE PRECISION :: STERRADIUS, STERHEIGHT, TENSION, DNARAD
  INTEGER :: MAXBENDINTER

  ! disco potential parameters
  INTEGER, PARAMETER :: MAXNTAIL = 10, MAXTAILLEN=20
  DOUBLE PRECISION :: LDEBYEINV, EPSEV, EPSEVT, SIGCC, SIGTC, SIGTL, SIGTT,SIGCL
  INTEGER :: NCHARGE,NTAIL,TAILNBEAD(MAXNTAIL)
  DOUBLE PRECISION :: CHARGEPOS(MAXNCHARGE,3), CHARGEMAG(MAXNCHARGE)
  DOUBLE PRECISION :: TAILPOS(MAXNTAIL, 3), TAILBEADINFO(MAXNTAIL,MAXTAILLEN,8)
  LOGICAL :: DISCOPOTENTIAL, OPTFLEXTAILSONLY,USEFLEXTAILS, TAILSTARTSET
 
  ! langowski potential parameters
  LOGICAL :: LANGOWSKIPOT
  DOUBLE PRECISION :: SIG0, EPS0, LJX, LJXP

  ! ---------------
  ! stuff for optimization
  ! -----------------
  INTEGER :: OPTPRINTFREQ, NOPTRUN, DUMPSTEPS, MAXOPTSTEP
  LOGICAL :: DISCORAMP
  DOUBLE PRECISION :: DISCORANGE
  DOUBLE PRECISION :: INTERMOPTOL, OPTOL
  LOGICAL :: DUMPCURRENT, RESTARTDUMP
  CHARACTER*100 :: DUMPFILE
  DOUBLE PRECISION :: MAXEJUMP, MAXSTEPSIZE, STEPDCR, DGUESS
  INTEGER :: MAXDCR
  INTEGER :: DUMPNBEND
  INTEGER :: MAXOPTATTEMPT

  ! ---------------
  ! Stuff for basinhopping
  ! ----------------
  INTEGER :: SEGHOPEVERY, NBASINHOP
  LOGICAL :: FLIPNUCMC, READDATAFILE
  DOUBLE PRECISION :: FLIPNUCPROB
  DOUBLE PRECISION :: BHFACC, BHFSAME,BHTOL
  INTEGER :: BHCHECKRANGE, MAXHOPTRY
  DOUBLE PRECISION :: STARTRRANGE, STARTTRANGE, STARTARANGE
  DOUBLE PRECISION :: BHTEMP

  ! -------------------
  ! stuff for saving and parsing databases
  ! -------------------
  INTEGER :: NCONFIGSAVE, NCONFIGDUMP
  CHARACTER*100 :: DATAFILE, NEWDATAFILE
  DOUBLE PRECISION :: DBCONFIGMAXE
  LOGICAL :: DODBMINIMIZE,DODBSORT, ENERGYRECALC

  ! --------- Output / input-------------
  CHARACTER*100 :: OUTFILE
  LOGICAL :: PRINTNUCDIST, PRINTEPARTS, PRINTFIBERDIAM

  LOGICAL :: RESTARTFROMFILE, CHECKRESTART
  CHARACTER*100 :: RESTARTFILE
  INTEGER :: RESTARTBEADS

  LOGICAL :: REPLICATESTRUCT
  CHARACTER*100 :: REPLICFILE
  LOGICAL :: RANDSTART
  DOUBLE PRECISION :: RSRANGE

  ! mark specific points on a nucleosome
  INTEGER, PARAMETER :: MAXMARKPOINT = 1000
  DOUBLE PRECISION :: MARKPOINTS(MAXMARKPOINT,3)
  INTEGER :: NMARKPOINT
  CHARACTER*3 :: MARKPOINTNAME(MAXMARKPOINT)
  

END MODULE KEYS
